@@ -2,8 +2,11 @@ import datetime
 from datetime import timedelta
 from unittest import mock
 
+import django
 import pytz
-from azure.storage.blob import Blob, BlobPermissions, BlobProperties
+from azure.core.exceptions import ResourceNotFoundError
+from azure.storage.blob import BlobProperties
+from django.core.exceptions import SuspiciousOperation
 from django.core.files.base import ContentFile
 from django.test import TestCase, override_settings
 
@@ -14,10 +17,13 @@ class AzureStorageTest(TestCase):
 
     def setUp(self, *args):
         self.storage = azure_storage.AzureStorage()
-        self.storage._service = mock.MagicMock()
-        self.storage._custom_service = mock.MagicMock()
+        self.storage._client = mock.MagicMock()
         self.storage.overwrite_files = True
+        self.account_name = 'test'
+        self.account_key = 'key'
         self.container_name = 'test'
+        self.storage.azure_account = self.account_name
+        self.storage.account_key = self.account_key
         self.storage.azure_container = self.container_name
 
     def test_get_valid_path(self):
@@ -79,202 +85,196 @@ class AzureStorageTest(TestCase):
 
     def test_get_available_name(self):
         self.storage.overwrite_files = False
-        self.storage._service.exists.side_effect = [True, False]
+        client_mock = mock.MagicMock()
+        client_mock.get_blob_properties.side_effect = [True, ResourceNotFoundError]
+        self.storage._client.get_blob_client.return_value = client_mock
         name = self.storage.get_available_name('foo.txt')
         self.assertTrue(name.startswith('foo_'))
         self.assertTrue(name.endswith('.txt'))
         self.assertTrue(len(name) > len('foo.txt'))
-        self.assertEqual(self.storage._service.exists.call_count, 2)
+        self.assertEqual(client_mock.get_blob_properties.call_count, 2)
 
     def test_get_available_name_first(self):
         self.storage.overwrite_files = False
-        self.storage._service.exists.return_value = False
+        client_mock = mock.MagicMock()
+        client_mock.get_blob_properties.side_effect = [ResourceNotFoundError]
+        self.storage._client.get_blob_client.return_value = client_mock
         self.assertEqual(
             self.storage.get_available_name('foo bar baz.txt'),
             'foo bar baz.txt')
-        self.assertEqual(self.storage._service.exists.call_count, 1)
+        self.assertEqual(client_mock.get_blob_properties.call_count, 1)
 
     def test_get_available_name_max_len(self):
         self.storage.overwrite_files = False
         # if you wonder why this is, file-system
         # storage will raise when file name is too long as well,
         # the form should validate this
+        client_mock = mock.MagicMock()
+        client_mock.get_blob_properties.side_effect = [False, ResourceNotFoundError]
+        self.storage._client.get_blob_client.return_value = client_mock
         self.assertRaises(ValueError, self.storage.get_available_name, 'a' * 1025)
-        self.storage._service.exists.side_effect = [True, False]
         name = self.storage.get_available_name('a' * 1000, max_length=100)  # max_len == 1024
         self.assertEqual(len(name), 100)
         self.assertTrue('_' in name)
-        self.assertEqual(self.storage._service.exists.call_count, 2)
+        self.assertEqual(client_mock.get_blob_properties.call_count, 2)
 
     def test_get_available_invalid(self):
         self.storage.overwrite_files = False
-        self.storage._service.exists.return_value = False
-        self.assertRaises(ValueError, self.storage.get_available_name, "")
-        self.assertRaises(ValueError, self.storage.get_available_name, "/")
-        self.assertRaises(ValueError, self.storage.get_available_name, ".")
-        self.assertRaises(ValueError, self.storage.get_available_name, "///")
+        self.storage._client.get_blob_properties.return_value = False
+        if django.VERSION[:2] == (3, 0):
+            # Django 2.2.21 added this security fix:
+            # https://docs.djangoproject.com/en/3.2/releases/2.2.21/#cve-2021-31542-potential-directory-traversal-via-uploaded-files
+            # It raises SuspiciousOperation before we get to our ValueError.
+            # The fix wasn't applied to 3.0 (no longer in support), but was applied to 3.1 & 3.2.
+            self.assertRaises(ValueError, self.storage.get_available_name, "")
+            self.assertRaises(ValueError, self.storage.get_available_name, "/")
+            self.assertRaises(ValueError, self.storage.get_available_name, ".")
+            self.assertRaises(ValueError, self.storage.get_available_name, "///")
+        else:
+            self.assertRaises(SuspiciousOperation, self.storage.get_available_name, "")
+            self.assertRaises(SuspiciousOperation, self.storage.get_available_name, "/")
+            self.assertRaises(SuspiciousOperation, self.storage.get_available_name, ".")
+            self.assertRaises(SuspiciousOperation, self.storage.get_available_name, "///")
         self.assertRaises(ValueError, self.storage.get_available_name, "...")
 
     def test_url(self):
-        self.storage._custom_service.make_blob_url.return_value = 'ret_foo'
-        self.assertEqual(self.storage.url('some blob'), 'ret_foo')
-        self.storage._custom_service.make_blob_url.assert_called_once_with(
-            container_name=self.container_name,
-            blob_name='some%20blob',
-            protocol='https')
+        blob_mock = mock.MagicMock()
+        blob_mock.url = 'https://ret_foo.blob.core.windows.net/test/some%20blob'
+        self.storage._client.get_blob_client.return_value = blob_mock
+        self.assertEqual(self.storage.url('some blob'), blob_mock.url)
+        self.storage._client.get_blob_client.assert_called_once_with('some blob')
 
     def test_url_unsafe_chars(self):
-        self.storage.custom_service.make_blob_url.return_value = 'ret_foo'
-        self.assertEqual(self.storage.url('foo;?:@=&"<>#%{}|^~[]`bar/~!*()\''), 'ret_foo')
-        self.storage.custom_service.make_blob_url.assert_called_once_with(
-            container_name=self.container_name,
-            blob_name='foo%3B%3F%3A%40%3D%26%22%3C%3E%23%25%7B%7D%7C%5E~%5B%5D%60bar/~!*()\'',
-            protocol='https')
+        blob_mock = mock.MagicMock()
+        blob_mock.url = 'https://ret_foo.blob.core.windows.net/test/some%20blob'
+        self.storage._client.get_blob_client.return_value = blob_mock
+        self.assertEqual(
+            self.storage.url('foo;?:@=&"<>#%{}|^~[]`bar/~!*()\''), blob_mock.url)
+        self.storage.client.get_blob_client.assert_called_once_with(
+            'foo;?:@=&"<>#%{}|^~[]`bar/~!*()\'')
+
+    @mock.patch('storages.backends.azure_storage.generate_blob_sas')
+    def test_url_expire(self, generate_blob_sas_mocked):
+        generate_blob_sas_mocked.return_value = 'foo_token'
+        blob_mock = mock.MagicMock()
+        blob_mock.url = 'https://ret_foo.blob.core.windows.net/test/some%20blob'
+        self.storage._client.get_blob_client.return_value = blob_mock
+        self.storage.account_name = self.account_name
 
-    def test_url_expire(self):
         utc = pytz.timezone('UTC')
         fixed_time = utc.localize(datetime.datetime(2016, 11, 6, 4))
-        self.storage._custom_service.generate_blob_shared_access_signature.return_value = 'foo_token'
-        self.storage._custom_service.make_blob_url.return_value = 'ret_foo'
         with mock.patch('storages.backends.azure_storage.datetime') as d_mocked:
             d_mocked.utcnow.return_value = fixed_time
-            self.assertEqual(self.storage.url('some blob', 100), 'ret_foo')
-            self.storage._custom_service.generate_blob_shared_access_signature.assert_called_once_with(
+            self.assertEqual(
+                self.storage.url('some blob', 100),
+                'https://ret_foo.blob.core.windows.net/test/some%20blob')
+            generate_blob_sas_mocked.assert_called_once_with(
+                self.account_name,
                 self.container_name,
                 'some blob',
-                permission=BlobPermissions.READ,
+                account_key=self.account_key,
+                user_delegation_key=None,
+                permission=mock.ANY,
                 expiry=fixed_time + timedelta(seconds=100))
-            self.storage._custom_service.make_blob_url.assert_called_once_with(
-                container_name=self.container_name,
-                blob_name='some%20blob',
-                sas_token='foo_token',
-                protocol='https')
 
-    def test_blob_service_default_params(self):
+    @mock.patch('storages.backends.azure_storage.generate_blob_sas')
+    def test_url_expire_user_delegation_key(self, generate_blob_sas_mocked):
+        generate_blob_sas_mocked.return_value = 'foo_token'
+        blob_mock = mock.MagicMock()
+        blob_mock.url = 'https://ret_foo.blob.core.windows.net/test/some%20blob'
+        self.storage._client.get_blob_client.return_value = blob_mock
+        self.storage.account_name = self.account_name
+        service_client = mock.MagicMock()
+        self.storage._service_client = service_client
+        self.storage.token_credential = 'token_credential'
+
+        utc = pytz.timezone('UTC')
+        fixed_time = utc.localize(datetime.datetime(2016, 11, 6, 4))
+        with mock.patch('storages.backends.azure_storage.datetime') as d_mocked:
+            d_mocked.utcnow.return_value = fixed_time
+            service_client.get_user_delegation_key.return_value = 'user delegation key'
+            self.assertEqual(
+                self.storage.url('some blob', 100),
+                'https://ret_foo.blob.core.windows.net/test/some%20blob')
+            generate_blob_sas_mocked.assert_called_once_with(
+                self.account_name,
+                self.container_name,
+                'some blob',
+                account_key=self.account_key,
+                user_delegation_key='user delegation key',
+                permission=mock.ANY,
+                expiry=fixed_time + timedelta(seconds=100))
+
+    def test_container_client_default_params(self):
         storage = azure_storage.AzureStorage()
+        storage.account_name = self.account_name
         with mock.patch(
-                'storages.backends.azure_storage.BlockBlobService',
-                autospec=True) as c_mocked:
-            self.assertIsNotNone(storage.service)
-            c_mocked.assert_called_once_with(
-                account_name=None,
-                account_key=None,
-                sas_token=None,
-                is_emulated=False,
-                protocol='https',
-                custom_domain=None,
-                connection_string=None,
-                token_credential=None,
-                endpoint_suffix=None)
+                'storages.backends.azure_storage.BlobServiceClient',
+                autospec=True) as bsc_mocked:
+            client_mock = mock.MagicMock()
+            bsc_mocked.return_value.get_container_client.return_value = client_mock
+            self.assertEqual(storage.client, client_mock)
+            bsc_mocked.assert_called_once_with(
+                'https://test.blob.core.windows.net',
+                credential=None)
 
-    def test_blob_service_params_no_emulator(self):
-        """Should ignore custom domain when emulator is not used"""
+    def test_container_client_params_account_key(self):
         storage = azure_storage.AzureStorage()
-        storage.is_emulated = False
-        storage.custom_domain = 'foo_domain'
-        with mock.patch(
-                'storages.backends.azure_storage.BlockBlobService',
-                autospec=True) as c_mocked:
-            self.assertIsNotNone(storage.service)
-            c_mocked.assert_called_once_with(
-                account_name=None,
-                account_key=None,
-                sas_token=None,
-                is_emulated=False,
-                protocol='https',
-                custom_domain=None,
-                connection_string=None,
-                token_credential=None,
-                endpoint_suffix=None)
-
-    def test_blob_service_params(self):
-        storage = azure_storage.AzureStorage()
-        storage.is_emulated = True
-        storage.endpoint_suffix = 'foo_suffix'
         storage.account_name = 'foo_name'
-        storage.account_key = 'foo_key'
-        storage.sas_token = 'foo_token'
         storage.azure_ssl = True
         storage.custom_domain = 'foo_domain'
+        storage.account_key = 'foo_key'
+        with mock.patch(
+                'storages.backends.azure_storage.BlobServiceClient',
+                autospec=True) as bsc_mocked:
+            client_mock = mock.MagicMock()
+            bsc_mocked.return_value.get_container_client.return_value = client_mock
+            self.assertEqual(storage.client, client_mock)
+            bsc_mocked.assert_called_once_with(
+                'https://foo_domain',
+                credential={'account_name': 'foo_name', 'account_key': 'foo_key'})
+
+    def test_container_client_params_sas_token(self):
+        storage = azure_storage.AzureStorage()
+        storage.account_name = 'foo_name'
+        storage.azure_ssl = False
+        storage.custom_domain = 'foo_domain'
+        storage.sas_token = 'foo_token'
+        with mock.patch(
+                'storages.backends.azure_storage.BlobServiceClient',
+                autospec=True) as bsc_mocked:
+            client_mock = mock.MagicMock()
+            bsc_mocked.return_value.get_container_client.return_value = client_mock
+            self.assertEqual(storage.client, client_mock)
+            bsc_mocked.assert_called_once_with(
+                'http://foo_domain',
+                credential='foo_token')
+
+    def test_container_client_params_token_credential(self):
+        storage = azure_storage.AzureStorage()
+        storage.account_name = self.account_name
+        storage.token_credential = 'foo_cred'
+        with mock.patch(
+                'storages.backends.azure_storage.BlobServiceClient',
+                autospec=True) as bsc_mocked:
+            client_mock = mock.MagicMock()
+            bsc_mocked.return_value.get_container_client.return_value = client_mock
+            self.assertEqual(storage.client, client_mock)
+            bsc_mocked.assert_called_once_with(
+                'https://test.blob.core.windows.net',
+                credential='foo_cred')
+
+    def test_container_client_params_connection_string(self):
+        storage = azure_storage.AzureStorage()
+        storage.account_name = self.account_name
         storage.connection_string = 'foo_conn'
-        storage.token_credential = 'foo_cred'
         with mock.patch(
-                'storages.backends.azure_storage.BlockBlobService',
-                autospec=True) as c_mocked:
-            self.assertIsNotNone(storage.service)
-            c_mocked.assert_called_once_with(
-                account_name='foo_name',
-                account_key='foo_key',
-                sas_token='foo_token',
-                is_emulated=True,
-                protocol='https',
-                custom_domain='foo_domain',
-                connection_string='foo_conn',
-                token_credential='foo_cred',
-                endpoint_suffix='foo_suffix')
-
-    def test_blob_custom_service_default_params(self):
-        storage = azure_storage.AzureStorage()
-        with mock.patch(
-                'storages.backends.azure_storage.BlockBlobService',
-                autospec=True) as c_mocked:
-            self.assertIsNotNone(storage.custom_service)
-            c_mocked.assert_called_once_with(
-                account_name=None,
-                account_key=None,
-                sas_token=None,
-                is_emulated=False,
-                protocol='https',
-                custom_domain=None,
-                connection_string=None,
-                token_credential=None,
-                endpoint_suffix=None)
-
-    def test_blob_custom_service_params_no_emulator(self):
-        """Should pass custom domain when emulator is not used"""
-        storage = azure_storage.AzureStorage()
-        storage.is_emulated = False
-        storage.custom_domain = 'foo_domain'
-        with mock.patch(
-                'storages.backends.azure_storage.BlockBlobService',
-                autospec=True) as c_mocked:
-            self.assertIsNotNone(storage.custom_service)
-            c_mocked.assert_called_once_with(
-                account_name=None,
-                account_key=None,
-                sas_token=None,
-                is_emulated=False,
-                protocol='https',
-                custom_domain='foo_domain',
-                connection_string=None,
-                token_credential=None,
-                endpoint_suffix=None)
-
-    def test_blob_custom_service_params(self):
-        storage = azure_storage.AzureStorage()
-        storage.is_emulated = True
-        storage.endpoint_suffix = 'foo_suffix'
-        storage.account_name = 'foo_name'
-        storage.account_key = 'foo_key'
-        storage.sas_token = 'foo_token'
-        storage.azure_ssl = True
-        storage.custom_domain = 'foo_domain'
-        storage.custom_connection_string = 'foo_conn'
-        storage.token_credential = 'foo_cred'
-        with mock.patch(
-                'storages.backends.azure_storage.BlockBlobService',
-                autospec=True) as c_mocked:
-            self.assertIsNotNone(storage.custom_service)
-            c_mocked.assert_called_once_with(
-                account_name='foo_name',
-                account_key='foo_key',
-                sas_token='foo_token',
-                is_emulated=True,
-                protocol='https',
-                custom_domain='foo_domain',
-                connection_string='foo_conn',
-                token_credential='foo_cred',
-                endpoint_suffix='foo_suffix')
+                'storages.backends.azure_storage.BlobServiceClient.from_connection_string',
+                spec=azure_storage.BlobServiceClient.from_connection_string) as bsc_mocked:
+            client_mock = mock.MagicMock()
+            bsc_mocked.return_value.get_container_client.return_value = client_mock
+            self.assertEqual(storage.client, client_mock)
+            bsc_mocked.assert_called_once_with('foo_conn')
 
     # From boto3
 
@@ -287,13 +287,13 @@ class AzureStorageTest(TestCase):
         with mock.patch('storages.backends.azure_storage.ContentSettings') as c_mocked:
             c_mocked.return_value = 'content_settings_foo'
             self.assertEqual(self.storage.save(name, content), name)
-            self.storage._service.create_blob_from_stream.assert_called_once_with(
-                container_name=self.container_name,
-                blob_name=name,
-                stream=content.file,
+            self.storage._client.upload_blob.assert_called_once_with(
+                name,
+                content.file,
                 content_settings='content_settings_foo',
-                max_connections=2,
-                timeout=20)
+                max_concurrency=2,
+                timeout=20,
+                overwrite=True)
             c_mocked.assert_called_once_with(
                 content_type='text/plain',
                 content_encoding=None,
@@ -310,27 +310,25 @@ class AzureStorageTest(TestCase):
         file.write(content)
         written_file = file.file
         file.close()
-        self.storage._service.create_blob_from_stream.assert_called_once_with(
-            container_name=self.container_name,
-            blob_name=name,
-            stream=written_file,
+        self.storage._client.upload_blob.assert_called_once_with(
+            name,
+            written_file,
             content_settings=mock.ANY,
-            max_connections=2,
-            timeout=20)
+            max_concurrency=2,
+            timeout=20,
+            overwrite=True)
 
     def test_storage_exists(self):
-        self.storage._service.exists.return_value = True
         blob_name = "blob"
+        client_mock = mock.MagicMock()
+        self.storage._client.get_blob_client.return_value = client_mock
         self.assertTrue(self.storage.exists(blob_name))
-        self.storage._service.exists.assert_called_once_with(
-            self.container_name, blob_name, timeout=20)
+        self.assertEqual(client_mock.get_blob_properties.call_count, 1)
 
     def test_delete_blob(self):
         self.storage.delete("name")
-        self.storage._service.delete_blob.assert_called_once_with(
-            container_name=self.container_name,
-            blob_name="name",
-            timeout=20)
+        self.storage._client.delete_blob.assert_called_once_with(
+            "name", timeout=20)
 
     def test_storage_listdir_base(self):
         file_names = ["some/path/1.txt", "2.txt", "other/path/3.txt", "4.txt"]
@@ -340,11 +338,11 @@ class AzureStorageTest(TestCase):
             obj = mock.MagicMock()
             obj.name = p
             result.append(obj)
-        self.storage._service.list_blobs.return_value = iter(result)
+        self.storage._client.list_blobs.return_value = iter(result)
 
         dirs, files = self.storage.listdir("")
-        self.storage._service.list_blobs.assert_called_with(
-            self.container_name, prefix="", timeout=20)
+        self.storage._client.list_blobs.assert_called_with(
+            name_starts_with="", timeout=20)
 
         self.assertEqual(len(dirs), 2)
         for directory in ["some", "other"]:
@@ -366,11 +364,11 @@ class AzureStorageTest(TestCase):
             obj = mock.MagicMock()
             obj.name = p
             result.append(obj)
-        self.storage._service.list_blobs.return_value = iter(result)
+        self.storage._client.list_blobs.return_value = iter(result)
 
         dirs, files = self.storage.listdir("some/")
-        self.storage._service.list_blobs.assert_called_with(
-            self.container_name, prefix="some/", timeout=20)
+        self.storage._client.list_blobs.assert_called_with(
+            name_starts_with="some/", timeout=20)
 
         self.assertEqual(len(dirs), 1)
         self.assertTrue(
@@ -384,15 +382,17 @@ class AzureStorageTest(TestCase):
 
     def test_size_of_file(self):
         props = BlobProperties()
-        props.content_length = 12
-        self.storage._service.get_blob_properties.return_value = Blob(props=props)
+        props.size = 12
+        client_mock = mock.MagicMock()
+        client_mock.get_blob_properties.return_value = props
+        self.storage._client.get_blob_client.return_value = client_mock
         self.assertEqual(12, self.storage.size("name"))
 
     def test_last_modified_of_file(self):
         props = BlobProperties()
         accepted_time = datetime.datetime(2017, 5, 11, 8, 52, 4)
         props.last_modified = accepted_time
-        self.storage._service.get_blob_properties.return_value = Blob(props=props)
+        self.storage._client.get_blob_properties.return_value = props
         time = self.storage.modified_time("name")
         self.assertEqual(accepted_time, time)
 
