@@ -1,5 +1,6 @@
 import io
 from datetime import datetime
+from unittest import mock
 
 from django.core.exceptions import (
     ImproperlyConfigured, SuspiciousFileOperation,
@@ -11,12 +12,6 @@ from requests.models import Response
 
 from storages.backends import dropbox
 
-try:
-    from unittest import mock
-except ImportError:  # Python 3.2 and below
-    import mock
-
-
 FILE_DATE = datetime(2015, 8, 24, 15, 6, 41)
 FILE_METADATA_MOCK = mock.MagicMock(spec=FileMetadata)
 FILE_METADATA_MOCK.size = 4
